@@ -2,6 +2,12 @@
 
 use core::marker::PhantomData;
 
+#[cfg(feature = "alloc")]
+extern crate alloc;
+
+#[cfg(feature = "alloc")]
+use alloc::boxed::Box;
+
 #[cfg(not(feature = "riscv-ulp-hal"))]
 use esp_idf_sys::*;
 
@@ -117,6 +123,9 @@ pub trait TouchPin: Pin {
     fn touch_channel(&self) -> touch_pad_t;
 }
 
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+pub trait SubscribedPin: Pin {}
+
 pub struct Input;
 
 pub struct Output;
@@ -127,6 +136,122 @@ pub struct Disabled;
 
 pub struct Unknown;
 
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+pub struct SubscribedInput;
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+struct UnsafeCallback(*mut Box<dyn FnMut() + 'static>);
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl UnsafeCallback {
+    #[allow(clippy::type_complexity)]
+    pub fn from(boxed: &mut Box<Box<dyn FnMut() + 'static>>) -> Self {
+        Self(boxed.as_mut())
+    }
+
+    pub unsafe fn from_ptr(ptr: *mut c_types::c_void) -> Self {
+        Self(ptr as *mut _)
+    }
+
+    pub fn as_ptr(&self) -> *mut c_types::c_void {
+        self.0 as *mut _
+    }
+
+    pub unsafe fn call(&mut self) {
+        let reference = self.0.as_mut().unwrap();
+
+        (reference)();
+    }
+}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+static ISR_SERVICE_ENABLED: crate::mutex::Mutex<bool> = crate::mutex::Mutex::new(false);
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+unsafe extern "C" fn irq_handler(unsafe_callback: *mut esp_idf_sys::c_types::c_void) {
+    let mut unsafe_callback = UnsafeCallback::from_ptr(unsafe_callback);
+    unsafe_callback.call();
+}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+fn enable_isr_service() -> Result<(), EspError> {
+    let mut service_enabled = ISR_SERVICE_ENABLED.lock();
+    if !*service_enabled {
+        if let Err(e) = esp!(unsafe { esp_idf_sys::gpio_install_isr_service(0) }) {
+            return Err(e);
+        } else {
+            *service_enabled = true;
+        }
+    }
+    Ok(())
+}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+type ClosureBox = Box<Box<dyn FnMut()>>;
+
+/// The PinNotifySubscription represents the association between an InputPin and
+/// a registered isr handler.
+/// When the PinNotifySubscription is dropped, the isr handler is unregistered.
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+pub(crate) struct PinNotifySubscription(i32, ClosureBox);
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl PinNotifySubscription {
+    fn subscribe<P>(pin: &mut P, callback: impl FnMut() + 'static) -> Result<Self, EspError>
+    where
+        P: InputPin + Pin,
+    {
+        enable_isr_service()?;
+
+        let pin_number: i32 = pin.pin();
+
+        let callback: Box<dyn FnMut() + 'static> = Box::new(callback);
+        let mut callback = Box::new(callback);
+
+        let unsafe_callback = UnsafeCallback::from(&mut callback);
+
+        esp!(unsafe {
+            esp_idf_sys::gpio_isr_handler_add(
+                pin_number,
+                Some(irq_handler),
+                unsafe_callback.as_ptr(),
+            )
+        })?;
+
+        Ok(Self(pin_number, callback))
+    }
+}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl Drop for PinNotifySubscription {
+    fn drop(self: &mut PinNotifySubscription) {
+        esp!(unsafe { esp_idf_sys::gpio_isr_handler_remove(self.0) }).expect("Error unsubscribing");
+    }
+}
+
+/// Interrupt types
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+pub enum InterruptType {
+    PosEdge,
+    NegEdge,
+    AnyEdge,
+    LowLevel,
+    HighLevel,
+}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl From<InterruptType> for gpio_int_type_t {
+    fn from(interrupt_type: InterruptType) -> gpio_int_type_t {
+        match interrupt_type {
+            InterruptType::PosEdge => gpio_int_type_t_GPIO_INTR_POSEDGE,
+            InterruptType::NegEdge => gpio_int_type_t_GPIO_INTR_NEGEDGE,
+            InterruptType::AnyEdge => gpio_int_type_t_GPIO_INTR_ANYEDGE,
+            InterruptType::LowLevel => gpio_int_type_t_GPIO_INTR_LOW_LEVEL,
+            InterruptType::HighLevel => gpio_int_type_t_GPIO_INTR_HIGH_LEVEL,
+        }
+    }
+}
+
 /// Drive strength (values are approximates)
 #[cfg(not(feature = "riscv-ulp-hal"))]
 pub enum DriveStrength {
@@ -171,7 +296,13 @@ macro_rules! impl_base {
         {
             fn reset(&mut self) -> Result<(), EspError> {
                 #[cfg(not(feature = "riscv-ulp-hal"))]
-                let res = esp_result!(unsafe { gpio_reset_pin(self.pin()) }, ());
+                let res = {
+                    esp!(unsafe { gpio_reset_pin(self.pin()) })?;
+                    #[cfg(feature = "alloc")]
+                    self.disable_interrupt()?;
+                    Ok(())
+                };
+
                 #[cfg(feature = "riscv-ulp-hal")]
                 let res = Ok(());
 
@@ -269,6 +400,32 @@ macro_rules! impl_base {
 
                 Ok(())
             }
+
+            #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+            fn enable_interrupt(&mut self) -> Result<(), EspError> {
+                esp!(unsafe { gpio_intr_enable(self.pin()) })?;
+
+                Ok(())
+            }
+
+            #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+            fn disable_interrupt(&mut self) -> Result<(), EspError> {
+                esp!(unsafe { gpio_intr_disable(self.pin()) })?;
+                esp!(unsafe { gpio_set_intr_type(self.pin(), gpio_int_type_t_GPIO_INTR_DISABLE) })?;
+                unsafe { unregister_irq_handler(self.pin() as usize) };
+
+                Ok(())
+            }
+
+            #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+            fn set_interrupt_type(
+                &mut self,
+                interrupt_type: InterruptType,
+            ) -> Result<(), EspError> {
+                esp!(unsafe { gpio_set_intr_type(self.pin(), interrupt_type.into()) })?;
+
+                Ok(())
+            }
         }
 
         impl<MODE> embedded_hal::digital::ErrorType for $pxi<MODE> {
@@ -315,6 +472,29 @@ macro_rules! impl_pull {
     };
 }
 
+/// # Safety
+///
+/// - Access to `IRQ_HANDLERS` is not guarded because we only call register
+///   from the context of Pin which is owned and in the rights state
+///
+// Clippy in the CI seems to wrongfully catch a only_used_in_recursion
+// lint error in this function. We'll ignore it until it's fixed.
+#[allow(clippy::only_used_in_recursion)]
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+unsafe fn register_irq_handler(pin_number: usize, p: PinNotifySubscription) {
+    chip::IRQ_HANDLERS[pin_number] = Some(p);
+}
+
+/// # Safety
+///
+/// - Access to `IRQ_HANDLERS` is not guarded because we only call register
+///   from the context of Pin which is owned and in the rights state
+///
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+unsafe fn unregister_irq_handler(pin_number: usize) {
+    chip::IRQ_HANDLERS[pin_number].take();
+}
+
 macro_rules! impl_input_base {
     ($pxi:ident: $pin:expr) => {
         pub struct $pxi<MODE> {
@@ -355,6 +535,29 @@ macro_rules! impl_input_base {
             }
         }
 
+        #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+        impl $pxi<Input> {
+            /// # Safety
+            ///
+            /// The callback passed to this method is executed in the context of an
+            /// interrupt handler. So you should take care of what is done in it.
+            pub unsafe fn into_subscribed(
+                mut self,
+                callback: impl FnMut() + 'static,
+                interrupt_type: InterruptType,
+            ) -> Result<$pxi<SubscribedInput>, EspError> {
+                self.enable_interrupt()?;
+
+                self.set_interrupt_type(interrupt_type)?;
+
+                let callback = PinNotifySubscription::subscribe(&mut self, callback)?;
+
+                register_irq_handler(self.pin() as usize, callback);
+
+                Ok($pxi { _mode: PhantomData })
+            }
+        }
+
         impl<MODE> Pin for $pxi<MODE>
         where
             MODE: Send,
@@ -371,6 +574,9 @@ macro_rules! impl_input_base {
 
         impl_base!($pxi);
         impl_hal_input_pin!($pxi: Input);
+
+        #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+        impl_hal_input_pin!($pxi: SubscribedInput);
     };
 }
 
@@ -776,6 +982,12 @@ where
 
 impl InputPin for GpioPin<Input> {}
 
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl InputPin for GpioPin<SubscribedInput> {}
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl SubscribedPin for GpioPin<SubscribedInput> {}
+
 impl OutputPin for GpioPin<Output> {}
 
 impl InputPin for GpioPin<InputOutput> {}
@@ -785,6 +997,10 @@ impl OutputPin for GpioPin<InputOutput> {}
 impl_base!(GpioPin);
 impl_hal_input_pin!(GpioPin: Input);
 impl_hal_input_pin!(GpioPin: InputOutput);
+
+#[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+impl_hal_input_pin!(GpioPin: SubscribedInput);
+
 impl_hal_output_pin!(GpioPin: InputOutput);
 impl_hal_output_pin!(GpioPin: Output);
 
@@ -800,6 +1016,13 @@ mod chip {
     #[cfg(feature = "riscv-ulp-hal")]
     use crate::riscv_ulp_hal::sys::*;
 
+    #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+    pub(crate) static mut IRQ_HANDLERS: [Option<PinNotifySubscription>; 40] = [
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None, None, None, None, None, None, None,
+    ];
+
     // NOTE: Gpio26 - Gpio32 are used by SPI0/SPI1 for external PSRAM/SPI Flash and
     //       are not recommended for other uses
     pin!(Gpio0:0, IO, RTC:11, ADC2:1, NODAC:0, TOUCH:1);
@@ -977,6 +1200,14 @@ mod chip {
 
     use super::*;
 
+    #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+    pub(crate) static mut IRQ_HANDLERS: [Option<PinNotifySubscription>; 49] = [
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None,
+    ];
+
     // NOTE: Gpio26 - Gpio32 (and Gpio33 - Gpio37 if using Octal RAM/Flash) are used
     //       by SPI0/SPI1 for external PSRAM/SPI Flash and are not recommended for
     //       other uses
@@ -1217,6 +1448,12 @@ mod chip {
 
     use super::*;
 
+    #[cfg(feature = "alloc")]
+    pub(crate) static mut IRQ_HANDLERS: [Option<PinNotifySubscription>; 22] = [
+        None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
+        None, None, None, None, None, None, None,
+    ];
+
     // NOTE: Gpio12 - Gpio17 are used by SPI0/SPI1 for external PSRAM/SPI Flash and
     //       are not recommended for other uses
     pin!(Gpio0:0,   IO,   RTC:0,  ADC1:0, NODAC:0, NOTOUCH:0);
